/* $Id: e2_fileview.h 1396 2005-11-04 08:08:08Z tpgww $
 *
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef E2_FILEVIEW_H
#define E2_FILEVIEW_H

#include "emelfm2.h"
#include "e2_pane.h"

//filepane columns enumerator
enum { FILENAME = 0, SIZE, PERM, OWNER, GROUP,
	MODIFIED, ACCESSED, CHANGED, MAX_COLUMNS };

typedef enum { GT, LT, EQ } Operator;

typedef struct _FileView
{
	GtkListStore *store;
	GtkWidget *treeview;
	GtkTreeSelection *selection;	//pointer to selection for this view's treeview
	/* unless we want an on-going hash of listed items, it's faster
 		to quick-store the selected-item names
	GList *selected_names;  //slist of names of seelcted items (for reselection), or NULL */
	GHashTable *selected_names;  //names of selected items (for reselection), or NULL
	gint sort_column;  //copy of current sort column, can be <0
	GtkSortType sort_order; //copy of current sort order
	GtkWidget *sort_arrows[MAX_COLUMNS];
	GtkLabel *name_label;	//for showing active pane by changing label content
	gchar dir[PATH_MAX];	//path shown in this view (utf-8)
//#ifndef E2_FAM
	time_t dir_mtime;
	time_t dir_ctime;
//#endif
	GList *dir_history;  //list of structs with focused line etc for each opened dir
//	GtkWidget *filter_menu_item;  redundant
	GtkWidget *check_dirs; //ptr to filter sub-menu toggle-widget
	GtkWidget *check_name; //ptr to filter sub-menu toggle-widget
	struct
	{
		gchar pattern[NAME_MAX];
		gboolean invert_mask;
		gboolean case_sensitive;
		gboolean active;
	} filename_filter;
	GtkWidget *check_size;
	struct
	{
		off_t size;
		Operator op;
		gboolean active;
	} size_filter;
	GtkWidget *check_date;
	struct
	{
		time_t time;
		Operator op;
		enum {MTIME, ATIME, CTIME} time_type;
		gboolean active;
	} date_filter;

	gboolean filter_directories;
	gboolean show_hidden;  //whether to display hidden items in this view
	gint row;	//index of most-recently clicked row in the treeview
	gint drop_row;	//index of most-recently moused row during a drag operation
	gboolean lit;	//TRUE if drop-row is highlighted
	gboolean refresh; //flag for whether a view update is a refresh (T) or cd (F)
	gboolean completed; //flag to block recursive view updates
	gboolean blocked;	//flag to block view updates when refresh is disabled
	gchar last_find[NAME_MAX]; //buffer for storing names used in find task
#ifdef E2_LISTTHREADS
	gint toprow;	//top row to show when scrolling after screen re-build
	gint thumbx;	//more scrolling parameters
	gint upperx;
#endif
} FileView;

typedef struct _FileInfo
{
	gchar filename[NAME_MAX+1];  //utf-8 string
	struct stat statbuf;
} FileInfo;

typedef struct  _E2_SelectedItemInfo {
	gchar *filename;  //selected item name (except no trailing /), utf-8 string
#ifdef E2_INCLIST
	//use this if the store data is changed on the fly
	GtkTreeRowReference *ref;
#else
	//this is used for un-selecting liststore items on the fly
	GtkTreePath *path;
#endif
} E2_SelectedItemInfo;

/*typedef struct _E2_ViewItemData
{
	GtkTreePath *path;
	GtkTreeRowReference *ref;	
} E2_ViewItemData; */

FileView *curr_view;
FileView *other_view;

void e2_fileview_set_menu_position (GtkMenu *menu,
	gint *x, gint *y, gboolean *push_in, GtkWidget *treeview);
void e2_fileview_focus_row (FileView *view, gint row, 
	gboolean select_row, gboolean clear_selection,
	gboolean center, gboolean grab_focus);
//void e2_fileview_select_row_by_filename (FileView *view, gchar *filename); unused
GHashTable *e2_fileview_log_selected_names (FileView *view);
void e2_fileview_reselect_names (FileView *view);
void e2_fileview_translate_cols_array (gint *src_array, 
	gint *dest_array, gint size);
void e2_fileview_switch_views (void);
//void e2_fileview_switch_views_simple (void);
void e2_fileview_set_font (void);
void e2_fileview_set_row_background (FileView *view,
	GtkTreeIter *iter, GdkColor *color);
void e2_fileview_clear_row_background (FileView *view,
	GtkTreeIter *iter);
void e2_fileview_set_col_order (E2_PaneRuntime *rt);
void e2_fileview_update_col_cachedata (void);
void e2_fileview_get_scroll_data (FileView *view, gint *leftcol, gint *toprow);
void e2_fileview_scroll_to_position (FileView *view, gint leftcol, gint toprow);
//gint e2_fileview_find_name_col (FileView *view); unused
#ifdef E2_LISTTHREADS
FileView *
#else
void
#endif
  e2_fileview_prepare_list (FileView *view);
#ifdef E2_LISTTHREADS
void e2_fileview_threadedprepare_list (FileView *view);
#endif
void e2_fileview_scroll_list (GThread *viewthread);
void e2_fileview_refresh_list (FileView *view);
gboolean e2_fileview_treehash_free (GHashTable *data);
void e2_fileview_change_dir (FileView *view, gchar *path);
GtkWidget *e2_fileview_create_list (E2_PaneRuntime *rt);
void e2_fileview_select_all (GtkWidget *widget, FileView *view);
void e2_fileview_clean_selected (GPtrArray *selected);
GPtrArray *e2_fileview_get_selected (FileView *view);
GList *e2_fileview_get_selected_local (FileView *view);
FileInfo *e2_fileview_get_selected_first_local (FileView *view);
gchar *e2_fileview_get_row_name (FileView *view, gint row);

#endif //ndef E2_FILEVIEW_H
